// OpenGL graphics includes
#include <GL/glew.h>
#if defined(__APPLE__) || defined(MACOSX)
#include <GLUT/glut.h>
#else
#include <GL/freeglut.h>
#endif

// CUDA utilities and system includes
#include "include/cudawrapper.h"
#include <cuda_gl_interop.h>

//!!#include <cutil_inline.h>    // includes cuda.h and cuda_runtime_api.h
//!!#include <cutil_gl_inline.h>
//!!#include <cutil_gl_error.h>
//!!#include <cudaGL.h>

// Skeletonization APIs
#include "include/skelft.h"
#include "include/field.h"
#include "include/SimpleMesh.h"

#include <math.h>
#include <iostream>
#include <vector>
#include <algorithm>
#include <stack>


// constants / global variables
unsigned int window_width	= 512;
unsigned int window_height	= 512;
unsigned int image_width	= 512;
unsigned int image_height	= 512;
int			     fboSize		= 512;
const int    MAX_SKELPOINTS = 50000;
const int    min_skel_neighbors = 50;				//# skeleton neighbors (pixels) of the closest selected skel points from which we compute the rotation axis




// GUI and visualization controls
float		 rotateX = 0.0f, rotateY = 0.0f;
float		 translateX = 0.0f, translateY = 0.0f, translateZ = -5.0f;
int			 mouseOldX, mouseOldY;
int			 mouseButtons = 0;
int			 modifier = 0;
bool		 need_render = true;
bool		 show_skeleton_2d = true;
bool		 show_skeleton_3d = true;
bool		 show_mesh = true;
bool     display_skel_nbs = true;					  //When we show the rotation axis: do show the skeleton neighbors
bool     display_axis = true;               //When we show the rotation axis: do show the axis itself (line)
bool		 special = false;
bool		 core_component = true;							//Whether to keep only the core skeleton component or all the skeleton
bool		 show_axis = false;									//Whether to show the potential rotation axis before we click to initiate rotation
bool	   transparent = false;								//Whether to draw the scene with transparency/blending or not

float3	 closest;
int		   resolution = 5;										//Tesselation resolution of spheres used for rendering point clouds
float		 radius = 0.05;											//Radius of point-cloud spheres (as fraction of unit cube)
GLenum   cull_backface = GL_FRONT;					//Which faces are seen as backfaces in the model. Normally this is GL_FRONT, but if the triangles are specified in the
																						//wrong order (cw, not ccw), we need to flip this.

GLint		   viewport[4];											//Cached viewport values
GLdouble	 modelview[16];										//Cached modeview matrix
GLdouble	 projection[16];									//Cached projection matrix


// CUDA -> OpenGL interop:
GLuint pbo_dest;														//PBO holding the output of CUDA (skeletons)
cudaGraphicsResource *cuda_pbo_dest_resource;	//Handle to access the above PBO in CUDA
GLuint tex_cudaResult;												//GL texture bound to the above PBO (so we can draw it)

// OpenGL -> CUDA interop:
GLuint		framebuffer;											//Framebuffer for offscreen rendering
GLuint		depth_buffer;											//Depth buffer for offscreen rendering
GLuint		tex_screen;												//GL texture bound to the above
cudaGraphicsResource *cuda_tex_screen_resource;						//Handle to access the framebuffer in CUDA
int			   iGLUTWindowHandle = 0;						//Handle to the GLUT window

// Geometry:
GLuint		  meshGL = 0;										  //GL display-list for the mesh
SimpleMesh	mesh;													  //Holds the mesh we read and process


//Skeletonization stuff:
float*		siteParam;												//Boundary parameterization: fboSize x fboSize. value(i,j) = boundary-param of (i,j) if on boundary, else unused
unsigned char* outputSkeleton;							//Skeleton: fboSize x fboSize. value(i,j) = skeleton-presence at (i,j). Used to communicate CUDA<->CPU for various purposes.
float		  db[1024*1024];										//Near and far depth-buffers of the input shape from the current viewpoint
float		  db2[1024*1024];
float		  threshold = 2;										//Skeleton simplification value
short2		skelpoints[MAX_SKELPOINTS];				//2D skeleton points, saved as an array (for convenience)
float3		skelpoints_frame[MAX_SKELPOINTS];	//3D unprojected points corresponding to skelpoints[]
int			  num_skelpoints;										//Size of 'unprojected skelpoints' arrays
vector<float3> nbs;
float3      axis_center, axis_orient;
float2      axis_orient_2d;
float       axis_rotation = 0;
bool        axis_rotation_sign = true;
float       user_rotation[16] = {1,0,0,0,0,1,0,0,0,0,1,0,0,0,0,1};


// Various stuff
void runStdProgram(int argc, char** argv);
void cleanupNoPrompt();
void initCUDA( int argc, char **argv);
bool initGL(int *argc, char** argv);
void deletePBO(GLuint* pbo);
void createTextureSrc(GLuint* tex_screen, unsigned int size_x, unsigned int size_y);
void deleteTexture(GLuint* tex);
void createDepthBuffer(GLuint* depth, unsigned int size_x, unsigned int size_y);
void deleteDepthBuffer(GLuint* depth);
void createFramebuffer(GLuint* fbo, GLuint color, GLuint depth);
void deleteFramebuffer(GLuint* fbo);
template<typename T> void savePGM(T* data, int size, const char* fname);


// 2D skeleton functions
void floodNeighbors(unsigned char* image, int maxpts, int SZ, short2 seed, vector<float3>& result, vector<short2>& result2d);
void computeAxis(vector<float3>& nbs,vector<short2>& nbs2d,float3& axis_center,float3& axis_orient,float2& axis_orient_2d);

// GLUT callbacks
void display();
void keyboard(unsigned char key, int x, int y);
void reshape(int w, int h);
void passive_motion(int x, int y);


//----------------------------------------------------------------------------------------------------------------------------------



void renderMesh()
{
		vector<Point3d>& pts = mesh.points;													//Render the mesh (and cache it in the display-list)
		vector<face>&    fcs = mesh.faces;

		float* p_arr   = new float[pts.size()*3];									//We'll use vertex and normal arrays; seems a bit faster than drawing plain triangles
		float* n_arr   = new float[pts.size()*3];
		GLuint* id_arr = new GLuint[fcs.size()*3];
		int i=0;
		for(float* pp=p_arr, *nn=n_arr, *pe=pp+3*pts.size();pp<pe;++i)						//Copy vertices and normals in a format we can pass to GL
		{
			*pp++ = pts[i].x;
			*pp++ = pts[i].y;
			*pp++ = pts[i].z;
			*nn++ = pts[i].Normal()[0];
			*nn++ = pts[i].Normal()[1];
			*nn++ = pts[i].Normal()[2];
		}

		i=0;
		for(GLuint* ii=id_arr,*ie=ii+3*fcs.size();ii<ie;++i)					//Copy indices in a format we can pass to GL
		{
			*ii++ = fcs[i].index[0];
			*ii++ = fcs[i].index[1];
			*ii++ = fcs[i].index[2];
		}

		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_NORMAL_ARRAY);
		glVertexPointer(3, GL_FLOAT, 0, p_arr);												//Pass vertex/normal pointers to GL (this doesn't pass the actual data yet..)
		glNormalPointer(GL_FLOAT, 0, n_arr);

		meshGL = glGenLists(1);																				//Make display list for mesh
		glNewList(meshGL,GL_COMPILE_AND_EXECUTE);
		glDrawElements(GL_TRIANGLES, 3*fcs.size(), GL_UNSIGNED_INT, id_arr);				//Draw the triangles; this pulls the vertices/normals from the CPU to the GPU
		glEndList();

		delete[] p_arr;																								//Deallow temp buffers
		delete[] n_arr;
		delete[] id_arr;

		glDisableClientState(GL_VERTEX_ARRAY);												//Apparently we don't need arrays enabled any more, now all data is on the GPU
		glDisableClientState(GL_NORMAL_ARRAY);
}


void renderPointCloud()
{
    vector<Point3d>& pts = mesh.points;													//Render the mesh (and cache it in the display-list)

    meshGL = glGenLists(1);																			//Make display list for mesh
    glNewList(meshGL,GL_COMPILE_AND_EXECUTE);
    for(int i=0;i<pts.size();++i)
    {
          const Point3d& p = pts[i];
          glTranslatef(p.x,p.y,p.z);
          glutSolidSphere(radius,resolution,resolution);
          glTranslatef(-p.x,-p.y,-p.z);
    }
    glEndList();
}


void renderScene(bool full)																	//Render the 3D shape
{
	float lDiff[] = {0,0.8,1.0};
	float lSpec[] = {0.5,0.5,0.5};
  float spec = 128;

	if (full) glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	else glClear(GL_DEPTH_BUFFER_BIT);

  glMatrixMode(GL_MODELVIEW);																//Go to the current viewpoint

  glLoadIdentity();
  glTranslatef(translateX, translateY, translateZ);										// -add up the user-specified view parameters
  glRotatef(rotateX, 1.0, 0.0, 0.0);
  glRotatef(rotateY, 0.0, 1.0, 0.0);
	glMultMatrixf(user_rotation);

	glTranslatef(axis_center.x,axis_center.y,axis_center.z);								//Apply user-rotation around rotation axis thru rotation center
	glRotatef(axis_rotation,axis_orient.x,axis_orient.y,axis_orient.z);
	glTranslatef(-axis_center.x,-axis_center.y,-axis_center.z);

	glGetIntegerv(GL_VIEWPORT, viewport);
	glGetDoublev(GL_MODELVIEW_MATRIX, modelview);
	glGetDoublev(GL_PROJECTION_MATRIX, projection);

  if (full)
	{
		glEnable(GL_LIGHTING);
		glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,  lDiff);
		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, lSpec);
		glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 128);
	}
	else glDisable(GL_LIGHTING);
  glEnable(GL_DEPTH_TEST);																//Depth test must be enabled in any case, since we require the depth buffer and/or the correctly depth-hidden scene


  if (transparent)
	{
	  glColor4f(lDiff[0],lDiff[1],lDiff[2],0.1);
	  glEnable(GL_BLEND);
	  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	  glEnable(GL_COLOR_MATERIAL);
    glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
  }

	if (!meshGL)																				  	//Build  display-list if we don't have it already
	{
		if (mesh.faces.size())
	  	    renderMesh();
		else
		    renderPointCloud();
	}
	else
		glCallList(meshGL);																		//Draw existing display-list if we have one

  glDisable(GL_BLEND);
	glDisable(GL_COLOR_MATERIAL);
  glDisable(GL_LIGHTING);
}



void floodFill(unsigned char* image, int SZ, short2 seed, int& num_skelpoints, short2* skelpoints)
//Flood fill on image[] from seed: mark visited pixels in image[], collect 2D skeleton points in skelpoints[]
{
	stack<short2> q;
	q.push(seed);
	image[seed.x+SZ*seed.y] = 128;												//Mark seed as 'visited'

	num_skelpoints = 0;																		//We start gathering a new set of 2D skelpoints

	while(!q.empty())																			//Do the flood-fill
	{
		short2 p = q.top(); q.pop();												//Get a seed; all seeds have a value of 128
		skelpoints[num_skelpoints] = p;											//This generates a new 2D skelpoint

		for(int i=p.x-1;i<=p.x+1;++i)												//Flood fill to seed's neighbors which are also on the skeleton
			for(int j=p.y-1;j<=p.y+1;++j)
			{
				if (i==p.x && j==p.y) continue;									//Skip seed itself
				if (image[i+SZ*j]!=255) continue;								//Neighbor is either background or visited, skip it
				image[i+SZ*j] = 128;														//Neighbot is skeleton and unvisited: mark it as 'visited'
				q.push(make_short2(i,j));												//...and add it to the flood fill
			}

		++num_skelpoints;
	}
}

void computeLargestComponent(short2 maxPos, int SZ, unsigned char* skel_cuda)
{
	//Copy skel_cuda (CUDA) -> outputSkeleton (CPU) for doing flood fill on latter
  cudaMemcpy(outputSkeleton, skel_cuda, SZ*SZ, cudaMemcpyDeviceToHost);

	//Do flood fill: keep skeleton component passing through maxPos, eliminate rest
	floodFill(outputSkeleton,fboSize,maxPos,num_skelpoints,skelpoints);

	//Pass to CUDA the skeleton in which the above connected component is 128, all other components are 255, and background is 0
  cudaMemcpy(skel_cuda, outputSkeleton, SZ*SZ, cudaMemcpyHostToDevice);

	skelft2DFillHoles(skel_cuda);

  //Unproject these 2D skel points to 3D.
	for(int i=0;i<num_skelpoints;++i)
	{
		  int winX = skelpoints[i].x, winY = skelpoints[i].y; float winZ;
		  double posX, posY, posZ, posX2, posY2, posZ2;

		  winZ = db2[winX + image_width*winY];												//Get one possible unprojection using the far depth-value in the rendered image
		  gluUnProject(winX, winY, winZ, modelview, projection, viewport, &posX2, &posY2, &posZ2);

		  winZ = db[winX + image_width*winY];												//Get a 2nd possible unprojection using the near depth-value in the rendered image
		  gluUnProject(winX, winY, winZ, modelview, projection, viewport, &posX, &posY, &posZ);

		  float3 newp = make_float3((posX+posX2)/2,(posY+posY2)/2,(posZ+posZ2)/2);			//We unproject the point to the AVERAGE of the two possibilities above (why? I don't know any better)

		  skelpoints_frame[i] = newp;																	//Collect unprojected 3D skel points from this frame (for later use)
	}
}



void computeSkeleton()																						//Compute 2D skeleton of current image in the framebuffer
{
	if (!need_render) return;

	unsigned char* out_data;
  size_t num_bytes; int rlen;

	glGetIntegerv(GL_PACK_ROW_LENGTH,&rlen);												//Set the row-length of the framebuffer we read to be the same as the buffer in which we compute skeletons.
	glPixelStorei(GL_PACK_ROW_LENGTH,fboSize);											//If not, we cannot compute the boundary-encoding directly from the framebuffer info.
	glReadPixels(0,0,image_width,image_height,GL_DEPTH_COMPONENT,GL_UNSIGNED_BYTE,outputSkeleton);
	glPixelStorei(GL_PACK_ROW_LENGTH,rlen);													//Save/restore the row-length parameter, otherwise rendering stuff gets screwed up later

  for(int i=0;i<image_width;++i)																  //Erase a 2-pixel-thick border around the grabbed Z framebuffer. This is needed since the
	{																																//skeletonization code assumes that the foreground shape (to skeletonize) is surrounded in
		outputSkeleton[i] = 255;																			//all 4 directions by background pixels.
		outputSkeleton[image_width+i] = 255;
		outputSkeleton[(image_height-2)*image_width+i] = 255;
		outputSkeleton[(image_height-1)*image_width+i] = 255;
	}

	for(int i=0;i<image_height;++i)
	{
		outputSkeleton[i*image_width] = 255;
		outputSkeleton[i*image_width+1] = 255;
		outputSkeleton[i*image_width+image_width-1] = 255;
		outputSkeleton[i*image_width+image_width-2] = 255;
	}



	glReadPixels(0,0,image_width,image_height,GL_DEPTH_COMPONENT,GL_FLOAT,db);				//!!Also get the depth-buffer as float (for later unprojection stuff)

  cudaGraphicsMapResources(1, &cuda_pbo_dest_resource, 0);				//Get CUDA-side pointer to the PBO buffer. Needed so we can read that buffer back to the CPU
	cudaGraphicsResourceGetMappedPointer((void **)&out_data, &num_bytes, cuda_pbo_dest_resource);

	short xm=1,ym=1,xM=image_width,yM=image_height;
	float maxImp;
	short2 maxPos;
	float length = skelft2DMakeBoundary(outputSkeleton,xm,ym,xM,yM,siteParam,fboSize,254,false);  //1. Parameterize boundary of inflated shape into 'siteParam' for skeleton computation
  skelft2DFT(0,siteParam,xm,ym,xM,yM,fboSize);											//2. Compute FT of 'siteParam'
  skelft2DSkeleton(out_data,length,threshold,xm,ym,xM,yM,&maxImp,&maxPos);
																																		//3. Skeletonize the FT into the CUDA buffer 'out_data', which is bound to 'cuda_pbo_dest_resource'
	if (core_component) computeLargestComponent(maxPos,fboSize,out_data);
																																		//4. Compute skeleton connected-component passing through skeleton 'center' maxPos

  cudaGraphicsUnmapResources(1, &cuda_pbo_dest_resource, 0);				//5. ??

  glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, pbo_dest);								//Bind the GL 'reading' buffer to pbo_dest so we can read from there
  glBindTexture(GL_TEXTURE_2D, tex_cudaResult);											//Make tex_cudaResult the active texture
  glTexSubImage2D(GL_TEXTURE_2D,0,0,0,image_width,image_height,GL_LUMINANCE,GL_UNSIGNED_BYTE,0);	//Apparently, this transfers the pixel data into the GL texture
  glBindBuffer(GL_PIXEL_PACK_BUFFER_ARB, 0);
  glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, 0);

  need_render = false;
}


void computeAxis(vector<float3>& nbs,vector<short2>& nbs2d,float3& axis_center,float3& axis_orient,float2& axis_orient_2d)
//Compute 3D axis which best approximates the nbs[] 3D-points.
//Return the axis as center (axis_center), orientation (axis_orient), and
{
	axis_center = make_float3(0,0,0);
	axis_orient = make_float3(0,0,0);
	short2 axis_center_2d = make_short2(0,0);
	const int NP = nbs.size();

	for(int i=0;i<NP;++i)															//1. Compute center (point that axis passes through)
	{
		axis_center.x += nbs[i].x;
		axis_center.y += nbs[i].y;
		axis_center.z += nbs[i].z;
		axis_center_2d.x += nbs2d[i].x;
		axis_center_2d.y += nbs2d[i].y;
	}

	axis_center.x /= NP;
	axis_center.y /= NP;
	axis_center.z /= NP;
	axis_center_2d.x /= NP;
	axis_center_2d.y /= NP;

	float d2max=0; int imax = 0;
	for(int i=0;i<NP;++i)
	{
		float3 nb = nbs[i];
		float  d2 = (nb.x-axis_center.x)*(nb.x-axis_center.x)+(nb.y-axis_center.y)*(nb.y-axis_center.y)+(nb.z-axis_center.z)*(nb.z-axis_center.z);
		if (d2<=d2max) continue;
		d2max = d2; imax = i;
	}

	float3 vmax = make_float3(nbs[imax].x-axis_center.x,nbs[imax].y-axis_center.y,nbs[imax].z-axis_center.z);

	float2 vmax2 = make_float2(nbs2d[imax].x-axis_center_2d.x,nbs2d[imax].y-axis_center_2d.y);
	float sign = (vmax2.y>0)? 1:-1;

	vmax.x *= sign;
	vmax.y *= sign;
	vmax.z *= sign;

	for(int i=0;i<NP;++i)
	{
		float3 nb   = nbs[i];
		short2 nb2d = nbs2d[i];
		float3  v; short2 v2d;
		v.x = nb.x-axis_center.x;																		//Vector from i-th point to center (in 3D)
		v.y = nb.y-axis_center.y;
		v.z = nb.z-axis_center.z;
		v2d.x = nb2d.x - axis_center_2d.x;													//Vector from i-th point to center (in 2D)
		v2d.y = nb2d.y - axis_center_2d.y;
		float dp = v.x*vmax.x+v.y*vmax.y+v.z*vmax.z;								//Make sure that we sum vectors (i-th point,center) in the same sense
																																//(which sense that is, shouldn't matter)
		float orient = (dp>0)? 1:-1;
		axis_orient.x += v.x*orient;
		axis_orient.y += v.y*orient;
		axis_orient.z += v.z*orient;
		axis_orient_2d.x += v2d.x*orient;
		axis_orient_2d.y += v2d.y*orient;
	}

	float nrm = sqrt(d2max/(axis_orient.x*axis_orient.x+axis_orient.y*axis_orient.y+axis_orient.z*axis_orient.z));
	axis_orient.x *= nrm;															//REMARK: This normalization is just for next visualizing the axis.
	axis_orient.y *= nrm;															//For rotation purposes, this shouldn't matter.
	axis_orient.z *= nrm;

	nrm = sqrt(axis_orient_2d.x*axis_orient_2d.x+axis_orient_2d.y*axis_orient_2d.y);
	axis_orient_2d.x /= nrm;
	axis_orient_2d.y /= nrm;
}


void floodNeighbors(unsigned char* image, int maxpts, int SZ, short2 seed, vector<float3>& result, vector<short2>& result2d)
{																							//Flood fill on image[] from seed; mark visited pixels in image[]
	deque<short2> q;
	q.push_front(seed);
	result.clear();
	result2d.clear();
	vector<short2> res;
	vector<short2> visited;

	while(!q.empty() && res.size()<maxpts)													//Start flood-fill:
	{
		short2 p = *q.begin(); q.pop_front();												//Get a seed; all seeds have a value of 128
		res.push_back(p);

		for(int i=p.x-1;i<=p.x+1;++i)														//Flood fill to seed's neighbors
			for(int j=p.y-1;j<=p.y+1;++j)
			{
				if ((i==p.x && j==p.y) || image[i+SZ*j]==0) continue;						//Skip background and the point itself, these aren't neighbors

				bool found = false;
				for(deque<short2>::const_iterator it=q.begin();!found && it!=q.end();++it)
				   if (it->x==i && it->y==j) found=true;
			  if (found) continue;

				found = false;
				for(int k=0;!found && k<res.size();++k)
					if (res[k].x==i && res[k].y==j) found=true;
				if (!found) q.push_back(make_short2(i,j));
			}
	}

	for(int i=0;i<res.size();++i)															//We have now found the 2D skel-points close to seed. We want to find the corresponding 3D skel-points.
	{																						//Since there's no direct mapping from a pixel (i,j) to data stored in our skelpoints_frame[],
		short2 p = res[i];																	//we need to search. Note that the search should always find something, since skelpoints[] contains all
		for(int j=0;j<num_skelpoints;++j)													//2D skeleton points in the current image frame.
		{
			short2 q = skelpoints[j];
			if (p.x==q.x && p.y==q.y)
			{
				result.push_back(skelpoints_frame[j]);
				result2d.push_back(q);
				break;
			}
		}
	}
}

void computeAxis()
{
	short2 closest2d;
	int min_d2 = 1.0e8;																		//Find 3D skel point closest to x,y. For this, we (a) search the closest 2D skelpoint to the mouse location (x,y)
	for(int i=0;i<num_skelpoints;++i)											//and (b) return the corresponding unprojected (3D) skelpoint.
	{
		short2 p = skelpoints[i];
		p.y = window_height - p.y;
		int d2  = (p.x-mouseOldX)*(p.x-mouseOldX) + (p.y-mouseOldY)*(p.y-mouseOldY);
		if (d2>min_d2) continue;
		min_d2 = d2;
		closest = skelpoints_frame[i];
		closest2d = skelpoints[i];
	}

	vector<short2> nbs2d;
	floodNeighbors(outputSkeleton,min_skel_neighbors,fboSize,closest2d,nbs,nbs2d);				//Find connected skel-component passing through closest2d
																																				//and containing at most 'min_skel_neighbors' points
	computeAxis(nbs,nbs2d,axis_center,axis_orient,axis_orient_2d);				//Compute rotation axis (center + direction-vector)

	short2 axis_location;
	axis_location.x = closest2d.x-mouseOldX;									//Vector from mouse-pos to axis center (in 2D)
	axis_location.y = window_height-closest2d.y-mouseOldY;
	float cp = axis_orient_2d.x*axis_location.y-axis_orient_2d.y*axis_location.x;
	axis_rotation_sign = (cp<0);												//This computes the rotation sign around the axis.
																											//The idea is that we always want to rotate so that stuff
																											//on the side of the axis where the mouse is will be brought towards the viewpoint.
}


void displaySkeleton()															//Display image on the screen, atop the current framebuffer
{
  glViewport(0, 0, window_width, window_height);
  glDisable(GL_DEPTH_TEST);
  glDisable(GL_LIGHTING);

	if (show_skeleton_2d && !special)
	{
      glBindTexture(GL_TEXTURE_2D, tex_cudaResult);			//We render from tex_cudaResult
      glEnable(GL_TEXTURE_2D);													//Set GL params to render a 2D texture onto the window
      glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

      glMatrixMode(GL_PROJECTION);											//Reset the projection/viewpoint so we can draw 1-to-1 in the window
      glPushMatrix();
      glLoadIdentity();
      glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);
      glMatrixMode(GL_MODELVIEW);
 	    glPushMatrix();
      glLoadIdentity();

	    glEnable(GL_COLOR_LOGIC_OP);											//Draw skeleton 'transparent-like'
	    glLogicOp(GL_AND_INVERTED);												//(note that skel==255, non-skel==0)
	    glBegin(GL_QUADS);
      glTexCoord2f(0.0, 0.0); glVertex3f(-1.0, -1.0, 0.5);
      glTexCoord2f(1.0, 0.0); glVertex3f(1.0, -1.0, 0.5);
      glTexCoord2f(1.0, 1.0); glVertex3f(1.0, 1.0, 0.5);
      glTexCoord2f(0.0, 1.0); glVertex3f(-1.0, 1.0, 0.5);
      glEnd();
	    glDisable(GL_COLOR_LOGIC_OP);
	    glDisable(GL_TEXTURE_2D);

      glMatrixMode(GL_PROJECTION);
      glPopMatrix();
      glMatrixMode(GL_MODELVIEW);
      glPopMatrix();
	}

	if (show_skeleton_3d)																//Draw the 3D skeleton and rotation-axis
	{
	  if (special || show_axis)											    //Show all details of the rotation axis construction
	  {
			glColor4f(1,0,0,0.1);
	    glDisable(GL_DEPTH_TEST);
		  glMatrixMode(GL_MODELVIEW);
	    glEnable(GL_LIGHTING);
		  float lDiff[] = {1,1,0};
		  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lDiff);

      if (display_skel_nbs)
		  for(int j=0;j<nbs.size();++j)										//1. Draw the skeleton neighbor points
		  {
			  glPushMatrix();
			  glTranslatef(nbs[j].x,nbs[j].y,nbs[j].z);
			  glutSolidSphere(0.02,10,10);
			  glPopMatrix();
		  }

      if (display_axis)
			{
		     float lDiff2[] = {0.1,1,0.1};										//2. Draw the axis center (closest skel point)
		     glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lDiff2);
		     glPushMatrix();
		     glTranslatef(axis_center.x,axis_center.y,axis_center.z);
		     glutSolidSphere(0.02,10,10);
		     glPopMatrix();

		     float lDiff3[] = {1,0,0};												//3. Draw the rotation axis itself
		     glDisable(GL_LIGHTING);
		     glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lDiff3);
		     glColor3fv(lDiff3);
		     glLineWidth(3);
		     glBegin(GL_LINES);
		     glVertex3f(axis_center.x-axis_orient.x,axis_center.y-axis_orient.y,axis_center.z-axis_orient.z);
		     glVertex3f(axis_center.x+axis_orient.x,axis_center.y+axis_orient.y,axis_center.z+axis_orient.z);
		     glEnd();
		     glLineWidth(1);
		  }

	    glDisable(GL_LIGHTING);
	  }
	}
}


void display()																			 			//GLUT display callback
{
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, framebuffer);	//Draw the scene in 'framebuffer', not on the screen

  glCullFace(cull_backface);															//Render the scene and collect the far Z-buffer in 'db2'
	glEnable(GL_CULL_FACE);																	//The far Z-buffer is given by the nearest-to-viewer backfacing geometry
	glDepthFunc(GL_LESS);
	glClearDepth(1.0);
	renderScene(false);
	glReadPixels(0,0,image_width,image_height,GL_DEPTH_COMPONENT,GL_FLOAT,db2);
  glDisable(GL_CULL_FACE);

	glDepthFunc(GL_LESS);																		//Render the scene again and compute the 2D skeleton; also read near Z-buffer in 'db'
	glClearDepth(1.0);
	renderScene(false);
	computeSkeleton();

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);						//Draw next on the screen (for visualization purposes)
	if (show_mesh) renderScene(true);												//First, draw the geometry
	else glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
																													//(I'd like to avoid that i.e. simply copy the image in framebuffer into the actual on-screen buffer, for speed)
	displaySkeleton();																			//Next, draw the skeleton (2D and/or 3D variants)

  cudaDeviceSynchronize();															  //REMARK: I needed to add this call so the timing is consistent.
  glutSwapBuffers();
}


void print_help()
{
	cout<<"Mouse commands:"<<endl;
	cout<<" left-button click-and drag: rotate in trackball mode"<<endl;
	cout<<" mid-button click-and-drag: rotate in skeleton axis mode"<<endl<<endl;

	cout<<"Keyboard options:"<<endl;
	cout<<" s:     toggle showing the 2D skeleton"<<endl;
	cout<<" S:     toggle showing the 3D skeleton during rotation"<<endl;
  cout<<" m:     toggle showing the object (mesh)"<<endl;
	cout<<" space: toggle showing the 3D skeleton during mouse motion"<<endl;
	cout<<" n:     display skeleton neighbor-points (yellow balls)"<<endl;
	cout<<" a:     display skeleton-based rotation axis (red line)"<<endl;
	cout<<" f:     toggle culling mesh backfaces on/off"<<endl;
	cout<<" t:     toggle displaying the mesh transparent or not"<<endl;
	cout<<" +,-:   increase/decrease the 2D skeleton simplification"<<endl;
	cout<<" b,B:   decrease/increase ball tesselation (point-cloud rendering)"<<endl;
	cout<<" r,R:   decrease/increase ball radius (point-cloud rendering)"<<endl;
	cout<<" h:     print this help"<<endl;
	cout<<" ESC:   exit"<<endl;
}

void keyboard(unsigned char key, int x, int y)											//GLUT keyboard callback
{
  switch(key)
	{
	case 's':
		    show_skeleton_2d = !show_skeleton_2d;
		    break;
	case 'S':
		    show_skeleton_3d = !show_skeleton_3d;
		    break;
	case 'm':
		    show_mesh = !show_mesh;
		    break;
	case ' ':
	  		show_axis = !show_axis;
				if (show_axis) passive_motion(x,y);
				break;
	case 'n':
	 		  display_skel_nbs = !display_skel_nbs;
	 		  need_render = true;
				break;
	case 'a':
				display_axis = !display_axis;
			  need_render = true;
				break;
	case 'f':
				cull_backface = (cull_backface==GL_FRONT)? GL_BACK:GL_FRONT;
			  need_render = true;
				break;
	case 't':
				transparent = !transparent;
				break;
	case '=':
	case '+':
			 	threshold += 0.05;
			 	need_render = true;
			  break;
	case '-':
	case '_':
			 	if (threshold > 0.5) threshold -= 0.05;
			 	need_render = true;
			  break;
	case 'B':
	      resolution++;
	      break;
  case 'b':
        if (resolution>3) resolution--;
        break;
  case 'R':
        radius += 0.005;
        glDeleteLists(meshGL,1);
        meshGL = 0;
        break;
  case 'r':
        radius -= 0.005;
        glDeleteLists(meshGL,1);
        meshGL = 0;
        break;
	case 'c':
	   core_component = !core_component;
		 need_render = true;
		 break;
	case 'h':
	   print_help();
		 break;
  case(27) :
     cleanupNoPrompt();
     break;
  }

	glutPostRedisplay();
}


void motion(int x, int y)																	//Mouse motion callback (for interaction)
{
    float dx, dy;
    dx = (float)(x - mouseOldX);													//Compute how much we moved the mouse
    dy = (float)(y - mouseOldY);

    if (mouseButtons)																			//Don't do anything unless a mouse-button is pressed
	  {
			if (special)
			{
				   Point3d perp = Point3d(0,0,-1).cross(Point3d(axis_orient_2d.x,axis_orient_2d.y,0));
				   float rot_delta = (perp.x*dx+perp.y*dy)*0.2;
				   if (axis_rotation_sign) axis_rotation += rot_delta;
			     else axis_rotation += 360 - rot_delta;
			     if (axis_rotation>360) axis_rotation -= 360;
			}
			else																										//Normal trackball interaction
			{
		  if (modifier & GLUT_ACTIVE_ALT)													//Alt: pan
		  {
			  translateX += dx * 0.01f;
			  translateY -= dy * 0.01f;
		  }
		  else if (modifier & GLUT_ACTIVE_CTRL)										//Ctrl: zoom
		  {
			  translateZ += dy * 0.01f;
		  }
		  else																										//No modifier: rotate
		  {
			  rotateX += dy * 0.2f;
			  rotateY += dx * 0.2f;
		  }
		  }

		  need_render = true;
    }

    mouseOldX = x;
    mouseOldY = y;

		glutPostRedisplay();
}

void passive_motion(int x, int y)																	//Mouse passive motion callback (when mouse moved but no button pressed)
{
	  if (show_axis)
		{
			mouseOldX = x;
			mouseOldY = y;
			computeAxis();
			glutPostRedisplay();
		}
}


void mouse(int button, int state, int x, int y)										//Mouse click callback (for interaction)
{
    if (state==GLUT_DOWN)
	  {
        mouseButtons |= 1<<button;
    }
	  else if (state == GLUT_UP)
	  {
        mouseButtons = 0;
				modifier = 0;
		    if (special)
		    {
			    glMatrixMode(GL_MODELVIEW);												//Transfer current interactive-user-rotation (around current axis) into aggregated rotation. For this, we:
			    glPushMatrix();															//- apply the currently stored user-rotation
			    glLoadMatrixf(user_rotation);											//- apply the current interactive user-rotation (atop of previous)
			    glTranslatef(axis_center.x,axis_center.y,axis_center.z);				//- read the aggregated result and save it into the stored user-rotation.
			    glRotatef(axis_rotation,axis_orient.x,axis_orient.y,axis_orient.z);	    //This way, consecutive interactive user-rotations are nicely cumulated
			    glTranslatef(-axis_center.x,-axis_center.y,-axis_center.z);
			    glGetFloatv(GL_MODELVIEW_MATRIX,user_rotation);
			    glPopMatrix();

			    axis_rotation = 0;														//Reset the interactive user rotation so we have it zero when we rotate next time
			    special = false;
			    need_render = true;
		}
  }

	modifier = glutGetModifiers();													//Save modifiers (GLUT doesn't let us read them in the motion-cb)
  mouseOldX = x;
  mouseOldY = y;

	if (mouseButtons==2 && state==GLUT_DOWN)								//User wants to rotate around skel-axis closest to mouse position:
	{
		computeAxis();
		special = true;
	}

  glutPostRedisplay();
}


void reshape(int w, int h)
{
    window_width = w;
    window_height = h;
		fboSize = 512; //!! skelft2DSize(window_width,window_height);									//   Get size of the image that CUDA will actually use to process our nx x ny image
}



//--------- Various utilities ----------------------------------------------------------------


void createPBO(GLuint* pbo, struct cudaGraphicsResource **pbo_resource,						//Create PBO (GL buffer that'll be used to receive the CUDA computations)
			   GLuint* tex_cudaResult, unsigned int size_x, unsigned int size_y)				//Next, bind this PBO to a 2D texture, so we can draw the stuff in the buffer
{
    glGenBuffers(1, pbo);																													//Create buffer object
    glBindBuffer(GL_ARRAY_BUFFER, *pbo);																					//?
    glBufferData(GL_ARRAY_BUFFER, fboSize*fboSize, 0, GL_DYNAMIC_DRAW);						//Tell how big the buffer is and how it'll be used
    glBindBuffer(GL_ARRAY_BUFFER, 0);																							//?
    cudaGraphicsGLRegisterBuffer(pbo_resource, *pbo, cudaGraphicsMapFlagsNone);		//Register this buffer object with CUDA

    glGenTextures(1, tex_cudaResult);																							//Create a texture for receiving the skeletons computed on CUDA.
    glBindTexture(GL_TEXTURE_2D, *tex_cudaResult);																//This is a 2D 1-byte-per-pixel texture, since this is how we compute skeletons in CUDA.
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE, size_x, size_y, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, 0);
}

void deletePBO(GLuint* pbo)
{
    glDeleteBuffers(1, pbo);
    *pbo = 0;
}



void createTextureSrc(GLuint* tex_screen, unsigned int size_x, unsigned int size_y)
{
    glGenTextures(1, tex_screen);																//Create a texture
    glBindTexture(GL_TEXTURE_2D, *tex_screen);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA16F_ARB, size_x, size_y, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);

    // register this texture with CUDA
    cudaGraphicsGLRegisterImage(&cuda_tex_screen_resource, *tex_screen, GL_TEXTURE_2D, cudaGraphicsMapFlagsReadOnly);
}

void deleteTexture(GLuint* tex)
{
    glDeleteTextures(1, tex);
    *tex = 0;
}

void createDepthBuffer(GLuint* depth, unsigned int size_x, unsigned int size_y)
{
    glGenRenderbuffersEXT(1, depth);															//Create a render buffer
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, *depth);
    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24, size_x, size_y);		//Allocate storage
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0);
}


void deleteDepthBuffer(GLuint* depth)
{
    glDeleteRenderbuffersEXT(1, depth);
    *depth = 0;
}

void createFramebuffer(GLuint* fbo, GLuint color, GLuint depth)
{
    glGenFramebuffersEXT(1, fbo);																//Create a render buffer
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, *fbo);
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, color, 0);
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, depth);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}

void deleteFramebuffer( GLuint* fbo)
{
    glDeleteFramebuffersEXT(1, fbo);
    *fbo = 0;
}

void cleanupNoPrompt()
{
	cudaGraphicsUnregisterResource(cuda_tex_screen_resource);		// unregister this buffer object with CUDA
	cudaGraphicsUnregisterResource(cuda_pbo_dest_resource);
	deletePBO(&pbo_dest);
	deleteTexture(&tex_screen);
	deleteTexture(&tex_cudaResult);
	deleteDepthBuffer(&depth_buffer);
	deleteFramebuffer(&framebuffer);
	glDeleteLists(meshGL,1);
	cudaFreeHost(outputSkeleton);
	cudaFreeHost(siteParam);

  //ALEX cutilDeviceReset();
  if (iGLUTWindowHandle) glutDestroyWindow(iGLUTWindowHandle);

  exit(EXIT_SUCCESS);
}

void initGLBuffers()
{
    createPBO(&pbo_dest,&cuda_pbo_dest_resource,&tex_cudaResult,fboSize,fboSize);	//Create PBO and bind it to a tex which will receive CUDA result
    createTextureSrc(&tex_screen, image_width, image_height);						//Create texture for offscreen rendering
    createDepthBuffer(&depth_buffer, image_width, image_height);					//Create depth buffer for offscreen rendering
    createFramebuffer(&framebuffer, tex_screen, depth_buffer);						//Create framebuffer for offscreen rendering
}

int main(int argc, char** argv)																					//Entry point
{
	// First initialize OpenGL context, so we can properly set GL for CUDA.
  // Needed achieve optimal perf with OpenGL/CUDA interop.
  if(false == initGL(&argc, argv)) return 1;

  // Now initialize CUDA context (GL context has been created already)
  initCUDA(argc, argv);

    if(argc < 2)
    {
        cerr<<"Usage: "<<argv[0]<<" filename"<<endl;
				cerr<<"Filename can be:"<<endl;
				cerr<<"   PLY mesh file (.ply)"<<endl;
				cerr<<"   DMS point-cloud file (.dms)"<<endl;
        exit(1);
    }

    int nok;
    if (!strcmp(argv[1]+strlen(argv[1])-3,"ply"))
    {
       nok = mesh.readPly(argv[1]);
    }
    else if (!strcmp(argv[1]+strlen(argv[1])-3,"dms"))
    {
       nok = mesh.readDMS(argv[1]);
    }
    else nok = 1;

    if (nok)
    {
       cerr<<"Error: cannot read input file. Exiting."<<endl;
       exit(1);
    }

    cout<<"Read shape: "<<mesh.points.size()<<" points, "<<mesh.faces.size()<<" faces"<<endl;


    mesh.scaleDown();								//Scale mesh in [-1,1]
    mesh.setNormals();								//Compute mesh normals for rendering

    glutDisplayFunc(display);						//Register GLUT callbacks
    glutKeyboardFunc(keyboard);
    glutReshapeFunc(reshape);
    glutMouseFunc(mouse);
    glutMotionFunc(motion);
		glutPassiveMotionFunc(passive_motion);
    initGLBuffers();

    cout<<"Press 'h' for help"<<endl;

    glutMainLoop();									//Start the app

	  cleanupNoPrompt();

		return 0;
}

void initCUDA( int argc, char **argv)												//Initialize CUDA context
{
	cudaGLSetGLDevice(0);

	int size = 1024;
	skelft2DInitialization(size);
	cudaMallocHost((void**)&outputSkeleton,size*size*sizeof(unsigned char));
	cudaMallocHost((void**)&siteParam,size*size*sizeof(float));

}

template<typename T> void savePGM(T* data, int size, const char* fname)		//Save buffer as PGM image (for debugging)
{
	FILE* fp = fopen(fname,"wb");
	if (!fp) return;

	const int SIZE = 3000;
	unsigned char buf[SIZE];
	int bb=0;

	fprintf(fp,"P5 %d %d 255\n",size,size);
	for(const T* vend=data+size*size,*vptr=data;vptr<vend;vptr++)
	{
		float v = *vptr;
		buf[bb++] = (unsigned char)(int)(v);
		if (bb==SIZE)
		{  fwrite(buf,sizeof(unsigned char),SIZE,fp); bb = 0; }
	}
	if (bb) fwrite(buf,sizeof(unsigned char),bb,fp);

	fclose(fp);
}



bool initGL(int *argc, char **argv )												//Initialize OpenGL
{
    glutInit(argc, argv);															//Create GL context
    glutInitDisplayMode(GLUT_RGBA | GLUT_ALPHA | GLUT_DOUBLE | GLUT_DEPTH);
    glutInitWindowSize(window_width, window_height);
    iGLUTWindowHandle = glutCreateWindow("skelview");

    glewInit();																		//Initialize necessary OpenGL extensions
    if (! glewIsSupported(
        "GL_VERSION_2_0 "
        "GL_ARB_pixel_buffer_object "
        "GL_EXT_framebuffer_object "
        ))
				{
          cerr<<"ERROR: Support for necessary OpenGL extensions missing."<<endl;
          return false;
    		}

    //viewport
    glViewport(0, 0, window_width, window_height);

    //projection
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(40, (GLfloat)window_width / (GLfloat) window_height, 0.1, 100);
		//glOrtho(-2,2,-2,2,0.1,100);

		//shading
    glClearColor(1,1,1,1);
    glDisable(GL_DEPTH_TEST);
		glShadeModel(GL_SMOOTH);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,1);										//Needed so we can render open meshes
		glEnable(GL_LIGHT0);

  	return true;
}
