#include "include/ply2mesh.h"
#include <cstdlib>
#include <cstring>
#include <fstream>
#include <iostream>
#include <tr1/functional>

#include "ply/ply_parser.hpp"


using namespace std;

using namespace std::tr1::placeholders;




class ply_to_obj_converter
{
public:
  typedef int flags_type;
  enum { triangulate = 1 << 0 };
  ply_to_obj_converter(flags_type flags = 0);
  bool convert(std::istream& istream, const std::string& istream_filename,vector<Point3d> *points, vector<face> *faces);

private:
  void info_callback(const std::string& filename, std::size_t line_number, const std::string& message);
  void warning_callback(const std::string& filename, std::size_t line_number, const std::string& message);
  void error_callback(const std::string& filename, std::size_t line_number, const std::string& message);
  bool end_header_callback();

  std::tr1::tuple<std::tr1::function<void()>, std::tr1::function<void()> > element_definition_callback(const std::string& element_name, std::size_t count);
  template <typename ScalarType> std::tr1::function<void (ScalarType)> scalar_property_definition_callback(const std::string& element_name, const std::string& property_name);
  template <typename SizeType, typename ScalarType> std::tr1::tuple<std::tr1::function<void (SizeType)>, std::tr1::function<void (ScalarType)>, std::tr1::function<void ()> > list_property_definition_callback(const std::string& element_name, const std::string& property_name);
  void vertex_begin();
  void vertex_x(ply::float32 x);
  void vertex_y(ply::float32 y);
  void vertex_z(ply::float32 z);
  void vertex_end();
  void face_begin();
  void face_vertex_indices_begin(ply::uint8 size);
  void face_vertex_indices_element(ply::int32 vertex_index);
  void face_vertex_indices_end();
  void face_end();
  flags_type flags_;
  std::ostream* ostream_;
  double vertex_x_, vertex_y_, vertex_z_;
  std::size_t face_vertex_indices_element_index_, face_vertex_indices_first_element_, face_vertex_indices_previous_element_;

  vector<Point3d> *points;
  vector<face> *faces;
};

ply_to_obj_converter::ply_to_obj_converter(flags_type flags)
  : flags_(flags)
{
}

void ply_to_obj_converter::info_callback(const std::string& filename, std::size_t line_number, const std::string& message)
{
  std::cerr << filename << ":" << line_number << ": " << "info: " << message << std::endl;
}

void ply_to_obj_converter::warning_callback(const std::string& filename, std::size_t line_number, const std::string& message)
{
  std::cerr << filename << ":" << line_number << ": " << "warning: " << message << std::endl;
}

void ply_to_obj_converter::error_callback(const std::string& filename, std::size_t line_number, const std::string& message)
{
  std::cerr << filename << ":" << line_number << ": " << "error: " << message << std::endl;
}

std::tr1::tuple<std::tr1::function<void()>, std::tr1::function<void()> > ply_to_obj_converter::element_definition_callback(const std::string& element_name, std::size_t count)
{
  if (element_name == "vertex") {
    return std::tr1::tuple<std::tr1::function<void()>, std::tr1::function<void()> >(
      std::tr1::bind(&ply_to_obj_converter::vertex_begin, this),
      std::tr1::bind(&ply_to_obj_converter::vertex_end, this)
    );
  }
  else if (element_name == "face") {
    return std::tr1::tuple<std::tr1::function<void()>, std::tr1::function<void()> >(
      std::tr1::bind(&ply_to_obj_converter::face_begin, this),
      std::tr1::bind(&ply_to_obj_converter::face_end, this)
    );
  }
  else {
    return std::tr1::tuple<std::tr1::function<void()>, std::tr1::function<void()> >(0, 0);
  }
}

//template <>
//std::tr1::function<bool ()>
bool ply_to_obj_converter::end_header_callback()
{
  return true;
}


template <>
std::tr1::function<void (ply::float32)> ply_to_obj_converter::scalar_property_definition_callback(const std::string& element_name, const std::string& property_name)
{
  if (element_name == "vertex") {
    if (property_name == "x") {
      return std::tr1::bind(&ply_to_obj_converter::vertex_x, this, _1);
    }
    else if (property_name == "y") {
      return std::tr1::bind(&ply_to_obj_converter::vertex_y, this, _1);
    }
    else if (property_name == "z") {
      return std::tr1::bind(&ply_to_obj_converter::vertex_z, this, _1);
    }
    else {
      return 0;
    }
  }
  else {
    return 0;
  }
}

template <>
std::tr1::tuple<std::tr1::function<void (ply::uint8)>, std::tr1::function<void (ply::int32)>, std::tr1::function<void ()> > ply_to_obj_converter::list_property_definition_callback(const std::string& element_name, const std::string& property_name)
{
  if ((element_name == "face") && (property_name == "vertex_indices")) {
    return std::tr1::tuple<std::tr1::function<void (ply::uint8)>, std::tr1::function<void (ply::int32)>, std::tr1::function<void ()> >(
      std::tr1::bind(&ply_to_obj_converter::face_vertex_indices_begin, this, _1),
      std::tr1::bind(&ply_to_obj_converter::face_vertex_indices_element, this, _1),
      std::tr1::bind(&ply_to_obj_converter::face_vertex_indices_end, this)
    );
  }
  else {
    return std::tr1::tuple<std::tr1::function<void (ply::uint8)>, std::tr1::function<void (ply::int32)>, std::tr1::function<void ()> >(0, 0, 0);
  }
}

void ply_to_obj_converter::vertex_begin()
{
}

void ply_to_obj_converter::vertex_x(ply::float32 x)
{
  vertex_x_ = x;
}

void ply_to_obj_converter::vertex_y(ply::float32 y)
{
  vertex_y_ = y;
}

void ply_to_obj_converter::vertex_z(ply::float32 z)
{
  vertex_z_ = z;
}

void ply_to_obj_converter::vertex_end()
{
  //(*ostream_) << "v " << vertex_x_ << " " << vertex_y_ << " " << vertex_z_ << "\n";
  points->push_back(Point3d(vertex_x_, vertex_y_, vertex_z_));
}

void ply_to_obj_converter::face_begin()
{
  if (!(flags_ & triangulate)) {
    ;//(*ostream_) << "f";
  }
}

void ply_to_obj_converter::face_vertex_indices_begin(ply::uint8 size)
{
  face_vertex_indices_element_index_ = 0;
}

void ply_to_obj_converter::face_vertex_indices_element(ply::int32 vertex_index)
{
  if (flags_ & triangulate) {
    if (face_vertex_indices_element_index_ == 0) {
      face_vertex_indices_first_element_ = vertex_index;
    }
    else if (face_vertex_indices_element_index_ == 1) {
      face_vertex_indices_previous_element_ = vertex_index;
    }
    else {
      //(*ostream_) << "f " << (face_vertex_indices_first_element_ + 1) << " " << (face_vertex_indices_previous_element_ + 1) << " " << (vertex_index + 1) << "\n";
      faces->push_back(face(face_vertex_indices_first_element_,
			    face_vertex_indices_previous_element_,
			    vertex_index));
			    

      face_vertex_indices_previous_element_ = vertex_index;
    }
    ++face_vertex_indices_element_index_;
  }
  else {
    ;//(*ostream_) << " " << (vertex_index + 1);
  }
}

void ply_to_obj_converter::face_vertex_indices_end()
{
  if (!(flags_ & triangulate)) {
    ;//(*ostream_) << "\n";
  }
}

void ply_to_obj_converter::face_end()
{
}

bool ply_to_obj_converter::convert(std::istream& istream, const std::string& istream_filename,
				   vector<Point3d> *points, vector<face> *faces)
{
  ply::ply_parser::flags_type ply_parser_flags = 0;
  ply::ply_parser ply_parser(ply_parser_flags);

  ply_parser.info_callback(std::tr1::bind(&ply_to_obj_converter::info_callback, this, std::tr1::ref(istream_filename), _1, _2));
  ply_parser.warning_callback(std::tr1::bind(&ply_to_obj_converter::warning_callback, this, std::tr1::ref(istream_filename), _1, _2));
  ply_parser.error_callback(std::tr1::bind(&ply_to_obj_converter::error_callback, this, std::tr1::ref(istream_filename), _1, _2));

  ply_parser.element_definition_callback(std::tr1::bind(&ply_to_obj_converter::element_definition_callback, this, _1, _2));

  ply::ply_parser::scalar_property_definition_callbacks_type scalar_property_definition_callbacks;
  //ALEX ply::ply_parser::scalar_property_definition_callbacks_type::at<ply::float32>(scalar_property_definition_callbacks) = std::tr1::bind(&ply_to_obj_converter::scalar_property_definition_callback<ply::float32>, this, _1, _2);
  
  scalar_property_definition_callbacks.get<ply::float32>() = std::tr1::bind(&ply_to_obj_converter::scalar_property_definition_callback<ply::float32>, this, _1, _2);
  
  ply_parser.scalar_property_definition_callbacks(scalar_property_definition_callbacks);

  ply::ply_parser::list_property_definition_callbacks_type list_property_definition_callbacks;
  //ALEX: ply::at<ply::uint8, ply::int32>(list_property_definition_callbacks) = std::tr1::bind(&ply_to_obj_converter::list_property_definition_callback<ply::uint8, ply::int32>, this, _1, _2);

 
  list_property_definition_callbacks.get<ply::uint8,ply::int32>() = std::tr1::bind(&ply_to_obj_converter::list_property_definition_callback<ply::uint8, ply::int32>, this, _1, _2);

  ply_parser.list_property_definition_callbacks(list_property_definition_callbacks);


  ply_parser.end_header_callback(std::tr1::bind(&ply_to_obj_converter::end_header_callback, this));

  this->points = points;
  this->faces = faces;

  return ply_parser.parse(istream);
}


int read_ply(const char *istream_filename,vector<Point3d>& points, vector<face>& faces)
{
  ply_to_obj_converter::flags_type ply_to_obj_converter_flags = 0;

  ply_to_obj_converter_flags |= ply_to_obj_converter::triangulate;

  ply_to_obj_converter ply_to_obj_converter(ply_to_obj_converter_flags);

  std::ifstream ifstream;

  ifstream.open(istream_filename);
  if (!ifstream.is_open())
  {
    std::cerr << "read_ply: " << istream_filename << ": " << "no such file or directory" << "\n";
    return 1;
  }

  ply_to_obj_converter.convert(ifstream, istream_filename, &points, &faces);

  return 0;
}
